
#include "opt/problemBase.h"

#include "opt/equation.h"

#include <cassert>
#include <iostream>

using std::cerr;
using std::endl;
using std::string;
using std::vector;

struct OptProblemBase::M_Data {

  M_Data(const string &n,
	 const vector<OptEquation*> &eqns,
	 int numVars,
	 const vector<double> &varLowerBounds,
	 const vector<double> &varUpperBounds,
	 const vector<double> &eqnLowerBounds,
	 const vector<double> &eqnUpperBounds) :
    name(n),
    equations(eqns),
    variableLowerBounds(varLowerBounds),
    variableUpperBounds(varUpperBounds),
    equationLowerBounds(eqnLowerBounds),
    equationUpperBounds(eqnUpperBounds)
  {
    assert((size_t)numVars == varLowerBounds.size());
    assert(varLowerBounds.size() == varUpperBounds.size());
    assert(eqnLowerBounds.size() == eqnUpperBounds.size());
    assert(false == varLowerBounds.empty());
    assert(false == eqnLowerBounds.empty());

    dummyHessianIndices[0] = 666;
    
    numLinearCoefficients = 0;
    sizeOfGradient = 0;
    
    totalNumberOfEquations = 0;
    size_t equationOffsetIndex = 0;
    size_t gradientOffsetIndex = 0;
    
    for (size_t i = 0; i < equations.size(); ++i) {
      
      gradientOffsetIndices.push_back(gradientOffsetIndex);
      equationOffsetIndices.push_back(equationOffsetIndex);
            
      for (size_t j = 0; 
	   j < equations[i]->GetNumberOfLinearCoefficients(); 
	   ++j) {

	linearCoefficients.push_back
	  (equations[i]->GetLinearCoefficients()[j]);

	linearCoefficientFunctionIndices.push_back
	  (equationOffsetIndex + 
	   equations[i]->GetLinearCoefficientFunctionIndices()[j]);
	
	linearCoefficientVariableIndices.push_back
	  (equations[i]->GetLinearCoefficientVariableIndices()[j]);

	numLinearCoefficients++;
      }

      for (size_t j = 0; 
	   j < equations[i]->GetGradientSize(); ++j) {
	
	gradientVariableIndices.push_back
	  (equations[i]->GetGradientVariableIndices()[j]);

	gradientFunctionIndices.push_back
	  (equationOffsetIndex + 
	   equations[i]->GetGradientFunctionIndices()[j]);

	sizeOfGradient++;
      }
      
      gradientOffsetIndex += equations[i]->GetGradientSize();
      equationOffsetIndex += equations[i]->GetRangeDimension();
      totalNumberOfEquations += equations[i]->GetRangeDimension();
    }

    assert(eqnLowerBounds.size() == totalNumberOfEquations);

    if (true == linearCoefficients.empty()) {
      linearCoefficients.push_back(0);
      assert(linearCoefficientFunctionIndices.empty());
      assert(linearCoefficientVariableIndices.empty());
      linearCoefficientFunctionIndices.push_back(0);
      linearCoefficientVariableIndices.push_back(0);
    }

    if (true == gradientFunctionIndices.empty()) {
      gradientFunctionIndices.push_back(0);
      gradientVariableIndices.push_back(0);
    }

    assert(equationLowerBounds.size() == totalNumberOfEquations);
  }

  ~M_Data() {

  }
  
  
  void SetState(const double *x) const {
    for (size_t i = 0; i < equations.size(); ++i) {
      equations[i]->SetState(x);
    }
  }

  void Evaluate(double *f) const {
    memset(f,0,sizeof(double)*totalNumberOfEquations);
    
    for (size_t i = 0; i < equations.size(); ++i) {
      equations[i]->Evaluate(equationOffsetIndices[i],f);
    }
  }

  void Gradient(double *G) const {
    memset(G,0,sizeof(double)*sizeOfGradient);

    for (size_t i = 0; i < equations.size(); ++i) {
      equations[i]->Gradient(gradientOffsetIndices[i],G);
    }
  }
  
  string name;
  vector<OptEquation*> equations;
  vector<double> variableLowerBounds;
  vector<double> variableUpperBounds;
  vector<double> equationLowerBounds;
  vector<double> equationUpperBounds;

  int dummyHessianIndices[1];

  vector<double> linearCoefficients;
  vector<int> linearCoefficientVariableIndices;
  vector<int> linearCoefficientFunctionIndices;
  
  vector<int> gradientVariableIndices;
  vector<int> gradientFunctionIndices;

  vector<size_t> equationOffsetIndices;
  
  size_t totalNumberOfEquations;

  size_t numLinearCoefficients;
  size_t sizeOfGradient;

  vector<size_t> gradientOffsetIndices;
};


OptProblemBase::OptProblemBase(const string &name,
			       const vector<OptEquation*> &equations,
			       int numberOfVariables,
			       const vector<double> &variableLowerBounds,
			       const vector<double> &variableUpperBounds,
			       const vector<double> &equationLowerBounds,
			       const vector<double> &equationUpperBounds) :
  m_data(new M_Data(name,
		    equations,
		    numberOfVariables,
		    variableLowerBounds,
		    variableUpperBounds,
		    equationLowerBounds,
		    equationUpperBounds))
{

}

OptProblemBase::~OptProblemBase()
{
  delete m_data;
}

int 
OptProblemBase::GetNumberOfEquations() const
{
  return m_data->totalNumberOfEquations;
}
  
int 
OptProblemBase::GetNumberOfVariables() const
{
  return m_data->variableLowerBounds.size();
}

int 
OptProblemBase::GetNumberOfLinearCoefficients() const
{
  return m_data->numLinearCoefficients;
}

int 
OptProblemBase::GetGradientSize() const
{
  return m_data->sizeOfGradient;
}
  
int 
OptProblemBase::GetHessianSize() const
{
  return 0;
}

const int *
OptProblemBase::GetLinearCoefficientFunctionIndices() const
{
  return &(m_data->linearCoefficientFunctionIndices[0]);
}
  
const int *
OptProblemBase::GetLinearCoefficientVariableIndices() const
{
  return &(m_data->linearCoefficientVariableIndices[0]);
}

const double *
OptProblemBase::GetLinearCoefficients() const
{
  return &(m_data->linearCoefficients[0]);
}

const int *
OptProblemBase::GetGradientFunctionIndices() const
{
  return &(m_data->gradientFunctionIndices[0]);
}
  
const int *
OptProblemBase::GetGradientVariableIndices() const
{
  return &(m_data->gradientVariableIndices[0]);
}
  
const int *
OptProblemBase::GetHessianFunctionIndices() const
{
  return m_data->dummyHessianIndices;
}

const int *
OptProblemBase::GetHessianVariableIndices() const
{
  return m_data->dummyHessianIndices;
}

const double *
OptProblemBase::GetVariableLowerBounds() const
{
  return &(m_data->variableLowerBounds[0]);
}

const double *
OptProblemBase::GetVariableUpperBounds() const
{
  return &(m_data->variableUpperBounds[0]);
}

const double *
OptProblemBase::GetEquationLowerBounds() const
{
  return &(m_data->equationLowerBounds[0]);
}

const double *
OptProblemBase::GetEquationUpperBounds() const
{
  return &(m_data->equationUpperBounds[0]);
}

const string &
OptProblemBase::GetName() const
{
  return m_data->name;
}

void 
OptProblemBase::SetState(const double *x)
{
  m_data->SetState(x);
}
  
void 
OptProblemBase::Evaluate(double *f) const
{
  m_data->Evaluate(f);
}

void 
OptProblemBase::Gradient(double *G) const
{
  m_data->Gradient(G);
}
