
#include "opt/snoptSolver.h"

#include "opt/config.h"
#include "opt/problem.h"

#include <cassert>
#include <iostream>
#include <string>

using std::cerr;
using std::endl;
using std::string;

#define CWLEN 8

#define MAX(a,b) (a)<(b)?(b):(a)


// tack on an underscore to the fortran name of the function
#define FORTRAN_NAME(_n) _n##_

#define BIGBND 1e30

// make a typedef for the gnarly function call back
typedef void (*SnoptfCallback)(// status
			       int *,
			       // n 
			       const int *,
			       // x
			       const double *,
			       // needf
			       const int *,
			       // nF
			       const int *,
			       // f
			       double *,
			       // needG
			       const int *,
			       // lenG
			       const int *,
			       // G
			       double *,
			       // needH
			       const int *,
			       // lenH
			       const int *,
			       // H
			       double *,
			       // fmul
			       double *,
			       // cu
			       char *,
			       // lencu
			       const int *,
			       // iu
			       int *,
			       // leniu
			       const int *,
			       // ru
			       double *,
			       // lenru
			       const int *,
			       // cwlen
			       unsigned int);
		 
extern "C" {
  // these are based on the snoptf calls found in
  // ~/svnsrc/snopt/src
  // 
  
  void FORTRAN_NAME(sninit)(// file descriptor to print messages
			    const int *iPrint,  
			    // file descriptor to print summary
			    const int *iSumm,   
			    // the rest are workspace variables
			    // to be shared with all snopt calls
			    char *cw,     
			    const int *lencw,   
			    int *iw, 
			    const int *leniw, 
			    double *rw, 
			    const int *lenrw,
			    unsigned int cwLen);

  void FORTRAN_NAME(snsetr)(const char *buffer,    // the name of the option
			    const double *rvalue,  // the value
			    const int *iPrint, 
			    const int *iSumm, 
			    int *inform,
			    char *cw, 
			    const int *lencw, 
			    int *iw, 
			    const int *leniw, 
			    double *rw, 
			    const int *lenrw,
			    unsigned int bufferLen,
			    unsigned int cwLen);

  void FORTRAN_NAME(snseti)(const char* buffer,    // the name of the option
			    const int* ivalue,     // the value
			    const int* iPrint, 
			    const int* iSumm,
			    int* inform, 
			    char *cw, 
			    const int *lencw, 
			    int *iw, 
			    const int *leniw, 
			    double *rw, 
			    const int *lenrw,
			    unsigned int bufferLength,
			    unsigned int cwLength);
  
  void FORTRAN_NAME(snset)(const char* buffer,     // the name of the option 
			   const int* iPrint, 
			   const int* iSumm,
			   int* inform, 
			   char *cw, 
			   const int *lencw, 
			   int *iw, 
			   const int *leniw, 
			   double *rw, 
			   const int *lenrw,
			   unsigned int bufferLen,
			   unsigned int cwLen);
  
  // documentation copied from snoptf.f
  /*     ON ENTRY:
   *
   *     Start   specifies how a starting basis (and certain other items)
   *             are to be obtained.
   *             start =  0 (Cold) means that Crash should be used to
   *                      choose an initial basis, unless a basis file is
   *                      given by reference in the Specs file to an 
   *                      Old basis file. 
   *             start =  1 (Basis file) means the same (but is more
   *                      meaningful in the latter case).
   *             start =  2 (Warm) means that a basis is already defined
   *                      in xstate and Fstate (probably from an earlier
   *                      call).
   *
   *     neF     is the number  of problem functions in F, including the 
   *             objective function (if any) and the linear
   *             and nonlinear constraints.  Simple upper and lower bound
   *             constraints on the variables should not be included in  F.
   *             neF > 0.
   *
   *     n       is the number of variables.
   *             n > 0.
   *
   *     neA     is the number of nonzero entries in A. 
   *             neA >= 0.
   *
   *     nxname  is the number of 8-character column (i.e., variable) names
   *             provided in the array xnames.  If nxname = 1,  then there
   *             are NO column names (generic names will be used in the
   *             printed solution).  Otherwise, nxname = n and every 
   *             column name must be provided.
   *     
   *     nFname  is the number of 8-character row (i.e., constraint and
   *             objective) names provided in the array Fnames.  
   *             If nFname = 1,  then there are NO row names (generic 
   *             names will be used in the printed solution).  Otherwise,
   *             nFname = neF and every row name must be provided.
   *
   *     ObjAdd  is a constant that will be added to the objective.
   *             Typically ObjAdd = 0.0d+0.
   *
   *     Prob    is an 8-character name for the problem, used in the
   *             output.  A blank name can be assigned if necessary.
   *
   *     xlow(n) are the lower bounds on x.
   *
   *     xupp(n) are the upper bounds on x.
   *
   *     xnames(nxname) is an character*8 array of names for each x(j).
   *             If nxname =  1, xnames is not used.  The printed solution
   *             will use generic names for the variables.
   *             If nxname = n, xnames(j) should contain an 8 character
   *             name of the jth variable (j = 1, n).
   *
   *     Flow(n) are the lower bounds on F.  If component F(ObjRow)
   *             is being optimized,  Flow(ObjRow) is ignored.
   *                     
   *     Fupp(n) are the upper bounds on F.  If component F(ObjRow)
   *             is being optimized,  Fupp(ObjRow) is ignored.
   *
   *     Fnames(nFname) is an character*8 array of names for each F(i).
   *             If nFname =  1, Fnames is not used.  The printed solution
   *             will use generic names for the objective and constraints.
   *             If nName = neF, Fnames(j) should contain an 8 character
   *             name of the jth constraint (j=1:neF).
   *
   *     xstate(n) sometimes contains a set of initial states for each
   *             variable x.  See the following NOTES.
   *
   *     x(n)    is a set of initial values for each variable  x.
   *
   *  NOTES:  1. If start = 0 (Cold) or 1 (Basis file) and an OLD BASIS
   *             file is to be input, xstate and x need not be set at all.
   *
   *          2. Otherwise, xstate(1:n) must be defined for a cold start.
   *             If nothing special is known about the problem, or if
   *             there is no wish to provide special information,
   *             you may set xstate(j) = 0, x(j) = 0.0d+0 for all j=1:n.
   *             All variables will be eligible for the initial basis.
   *        
   *             Less trivially, to say that variable j will probably
   *             be equal to one of its bounds,
   *             set xstate(j) = 4 and x(j) = bl(j)
   *             or  xstate(j) = 5 and x(j) = bu(j) as appropriate.
   *        
   *          3. For Cold starts with no basis file, a Crash procedure
   *             is used to select an initial basis.  The initial basis
   *             matrix will be triangular (ignoring certain small
   *             entries in each column).
   *             The values xstate(j) = 0, 1, 2, 3, 4, 5 have the following
   *             meaning:
   *                
   *             xstate(j)  State of variable j during Crash
   *             ---------  --------------------------------
   *             0, 1, 3    Eligible for the basis.  3 is given preference.
   *             2, 4, 5    Ignored.
   *        
   *             After Crash, xstate(j) = 2 entries are made superbasic.
   *             Other entries not selected for the basis are made
   *             nonbasic at the value x(j) if bl(j) <= x(j) <= bu(j),
   *             or at the value bl(j) or bu(j) closest to x(j).
   *
   *          4. For Warm starts, all of Fstate(1:m) is assumed to be
   *             set to the values 0, 1, 2 or 3 from some previous call.
   *        
   *     Fmul(neF) contains an estimate of the Lagrange multipliers
   *             (shadow prices) for the F- constraints.  They are used
   *             to define the Lagrangian for the first major iteration. 
   *             If nothing is known about Fmul, set
   *             Fmul(i) = 0.0d+0, i = 1:neF
   *
   *     ON EXIT:
   *
   *     xstate(n) is the final state vector for x:
   *
   *                hs(j)    State of variable j    Normal value of x(j)
   *
   *                  0      nonbasic               bl(j)
   *                  1      nonbasic               bu(j)
   *                  2      superbasic             Between bl(j) and bu(j)
   *                  3      basic                  ditto
   *
   *             Very occasionally there may be nonbasic variables for
   *             which x(j) lies strictly between its bounds.
   *             If nInf = 0, basic and superbasic variables may be outside
   *             their bounds by as much as the Feasibility tolerance.
   *             Note that if Scale is specified, the Feasibility tolerance
   *             applies to the variables of the SCALED problem. 
   *             In this case, the variables of the original problem may be
   *             as much as 0.1 outside their bounds, but this is unlikely
   *             unless the problem is very badly scaled.
   *
   *     x(n)    contains the final variables.
   *
   *     F(neF)  contains the final values of F.
   *
   *     xmul(neF) is the vector of Lagrange multipliers (shadow prices)
   *             for the variables constraints.
   *
   *     Fmul(neF) is the vector of Lagrange multipliers (shadow prices)
   *             for the general constraints.
   *
   *     inform  says what happened; see the User's Guide.
   *             A summary of possible values follows:
   *
   *             Inform   Meaning
   *
   *                0     Optimal solution found.
   *                1     The problem is infeasible.
   *                2     The problem is unbounded (or badly scaled).
   *                3     Too many iterations.
   *                5     The Superbasics limit is too small.
   *                4     Final solution is feasible, but final optimality 
   *                      could not quite be achieved,
   *                6     Subroutine usrfun requested termination by
   *                      returning mode < 0.
   *                7     Subroutine usrfun seems to be giving incorrect
   *                      gradients.
   *                8     Subroutine usrfun seems to be giving incorrect
   *                      gradients.
   *                9     The current point cannot be improved.
   *               10     Numerical error in trying to satisfy the linear
   *                      constraints (or the linearized nonlinear
   *                      constraints).  The basis is very ill-conditioned.
   *
   *               20     Not enough storage for the basis factorization.
   *               21     Error in basis package.
   *               22     The basis is singular after several attempts to
   *                      factorize it (and add slacks where necessary).
   *
   *               30     An OLD BASIS file had dimensions that did not
   *                      match the current problem.
   *               32     System error.  Wrong number of basic variables.
   *
   *               41     Not enough storage to hold SNOPT local variables.
   *               42     Not enough char*8  storage to solve the problem.
   *               43     Not enough integer storage to solve the problem.
   *               44     Not enough real    storage to solve the problem.
   *
   *     mincw   says how much character storage is needed to solve the
   *             problem.  If inform = 42, the work array cw(lencw) was 
   *             too small.  snopt may be called again with lencw suitably 
   *             larger than mincw.
   *
   *     miniw   says how much integer storage is needed to solve the
   *             problem.  If inform = 43, the work array iw(leniw) was too
   *             small.  snopt may be called again with leniw suitably 
   *             larger than miniw.  (The bigger the better, since it is
   *             not certain how much storage the basis factors need.)
   *
   *     minrw   says how much real storage is needed to solve the problem.
   *             If inform = 44, the work array rw(lenrw) was too small.
   *             (See the comments above for miniw.)
   *
   *     nS      is the final number of superbasics.
   *
   *     nInf    is the number of infeasibilities.
   *
   *     sInf    is the sum    of infeasibilities.
   *
   *     cu(lencu), iu(leniu), ru(lenru)  are character, integer and real
   *             arrays of USER workspace.  These arrays are available to
   *             pass data to the user-defined routine usrfun.
   *             If no workspace is required, you can either use dummy
   *             arrays for cu, iu and ru, or use cw, iw and rw
   *             (see below).
   *
   *     cw(lencw), iw(leniw), rw(lenrw)  are character*8, integer and real
   *             arrays of workspace used by snopt.
   *             lencw  should be about at least 500.
   *             leniw  should be about max( 500, 20(m+n) ) or larger.
   *             lenrw  should be about max( 500, 40(m+n) ) or larger.
   *
   *     SNOPT package maintained by Philip E. Gill,
   *     Dept of Mathematics, University of California, San Diego.
   *
   *     08 Nov 1998: First version based on the snopt of SNOPT 5.3-4.
   *     25 Aug 1999: for SNOPT Version 6.0.
   *     04 Nov 2001: LP's solved explicitly.
   *     04 Nov 2001: Current version of snoptf.
   *     ==================================================================
   */
  int FORTRAN_NAME(snoptf)(const int *start, 
			   const int *nef, 
			   const int *n, 
			   const int *nxname, 
			   const int *nfname, 
			   const double *objadd, 
			   const int *objrow,
			   const char *prob, 
			   SnoptfCallback usrfun, 
			   SnoptfCallback snlog, 
			   const int *iafun, 
			   const int *javar, 
			   const int *lena, 
			   const int *nea, 
			   const double *a, 
			   const int *igfun, 
			   const int *jgvar, 
			   const int *leng, 
			   const int *neg, 
			   const int *ihvar, 
			   const int *jhvar, 
			   const int *lenhu, 
			   const int *nehu, 
			   const double *xlow, 
			   const double *xupp, 
			   const char *xnames, 
			   const double *flow, 
			   const double *fupp, 
			   const char *fnames, 
			   double *x, 
			   int *xstate, 
			   double *xmul, 
			   double *f, 
			   int *fstate, 
			   double *fmul, 
			   int *inform, 
			   int *mincw, 
			   int *miniw, 
			   int *minrw, 
			   int *ns, 
			   int *ninf, 
			   double *sinf, 
			   char *cu, 
			   const int *lencu, 
			   int *iu, 
			   const int *leniu, 
			   double *ru, 
			   const int *lenru, 
			   char *cw, 
			   const int *lencw, 
			   int *iw, 
			   const int *leniw, 
			   double *rw, 
			   const int *lenrw, 
			   unsigned int prob_len, 
			   unsigned int xnames_len, 
			   unsigned int fnames_len, 
			   unsigned int cu_len, 
			   unsigned int cw_len);

  void FORTRAN_NAME(snmem)(const int *m, // num functions
			   const int *n, // num variables
			   const int *ne, // num non zero jacobians
			   const int *neGcon, // also non zero jacobians?
			   const int *nnCon, // number of nonlinear constraints
			   const int *nnJac, // size of jacobian
			   const int *nnObj, // number of variables?
			   int *mincw,
			   int *miniw,
			   int *minrw,
			   char *cw,
			   const int *lencw,
			   int *iw,
			   const int *leniw,
			   double *rw,
			   const int *lenrw,
			   unsigned int cwLen);
			   
			   
};

struct OptSnoptSolver::M_Data {

  M_Data() : iPrint(0),
	     iSumm(0),
	     lencw(0),
	     leniw(0),
	     lenrw(0),
	     cw(NULL),
	     iw(NULL),
	     rw(NULL),
	     inform(0)
  {

  }

  ~M_Data() {
    FreeWorkspace();
  }

  void FreeWorkspace() {
    if (NULL != cw) {
      delete [] cw;
    }
    if (NULL != iw) {
      delete [] iw;
    }
    if (NULL != rw) {
      delete [] rw;
    }
  }

  void AllocateWorkspace(int lcw,
			 int liw,
			 int lrw) {
    lencw = lcw;
    leniw = liw;
    lenrw = lrw;

    cw = new char[lencw*CWLEN];
    iw = new int[leniw];
    rw = new double[lenrw];
  }

  void CalculateMemoryRequirements(const OptProblem &problem) {
    
    int numEquations = problem.GetNumberOfEquations();
    int numVariables = problem.GetNumberOfVariables();
    int nm = numEquations + numVariables;

    int jacobianSize = problem.GetGradientSize();
    
    int mincw = 0;
    int miniw = 0;
    int minrw = 0;
    
    FORTRAN_NAME(snmem)(&numEquations,
			&numVariables,
			&jacobianSize,
			&jacobianSize,
			&numEquations,
			&jacobianSize,
			&numVariables,
			&mincw,
			&miniw,
			&minrw,
			cw,
			&lencw,
			iw,
			&leniw,
			rw,
			&lenrw,
			CWLEN);

    FreeWorkspace();

    // make it bigger than necessary. first of all, snopt's calculations
    // are incorrect. second of all if you have more memory than needed
    // it runs faster apparently
    //AllocateWorkspace(mincw,100*miniw+3000,200*minrw+3000);
    // these numbers come from Adnan's implementation...
    // so basically I'm ignoring what's calculated above...
    // BOOTLEG!!!
    AllocateWorkspace(mincw,1800*nm,4000*nm);
  }

  void Init() {
    FORTRAN_NAME(sninit)(&iPrint,
			 &iSumm,
			 cw,
			 &lencw, 
			 iw, 
			 &leniw, 
			 rw, 
			 &lenrw,
			 CWLEN);

    inform = 0;
  }

  void CheckInform() {
    if (0 != inform) {
      cerr << "snopt is complaining: " << inform << endl;
    }
  }

  void SetOption(const string &prop) {
    FORTRAN_NAME(snset)(prop.c_str(),
			&iPrint,
			&iSumm,
			&inform,
			cw,
			&lencw, 
			iw, 
			&leniw, 
			rw, 
			&lenrw,
			prop.size(),
			CWLEN);
    CheckInform();
  }

  void SetOption(const string &prop,
		 int value) {
    FORTRAN_NAME(snseti)(prop.c_str(),
			 &value,
			 &iPrint,
			 &iSumm,
			 &inform,
			 cw,
			 &lencw, 
			 iw, 
			 &leniw, 
			 rw, 
			 &lenrw,
			 prop.size(),
			 CWLEN);
    CheckInform();
  }

  void SetOption(const string &prop,
		 double value) {
    FORTRAN_NAME(snsetr)(prop.c_str(),
			 &value,
			 &iPrint,
			 &iSumm,
			 &inform,
			 cw,
			 &lencw, 
			 iw, 
			 &leniw, 
			 rw, 
			 &lenrw,
			 prop.size(),
			 CWLEN);
    CheckInform();
  }

  void SetConfigOptions(const OptConfig &config) {
    
    for (size_t i = 0; i < config.GetBinaryProperties().size(); ++i) {
      SetOption(config.GetBinaryProperties()[i]);
    }

    for (size_t i = 0; i < config.GetIntProperties().size(); ++i) {
      SetOption(config.GetIntProperties()[i].first,
		config.GetIntProperties()[i].second);
    }

    for (size_t i = 0; i < config.GetDoubleProperties().size(); ++i) {
      SetOption(config.GetDoubleProperties()[i].first,
		config.GetDoubleProperties()[i].second);
    }
  }

  bool CheckInformAfterSolve()
  {
    switch (inform) {
    case 0:
      break;
    case 1:
      cerr << "The problem is infeasible" << endl;
      break;
    case 2:
      cerr << "The problem is unbounded (or badly scaled)" << endl;
      break;
    case 3:
      cerr << "Too many iterations" << endl;
      break;
    case 4:
      cerr << "Feasible solution, but the required accuracies in the dual"
	   << " infeasibilities could not be achieved." << endl;
      break;
    case 5:
      cerr << "The superbasics limit is too small " << endl;
      break;
    case 6:
      cerr << "Termination has been requested by the user" << endl;
      break;
    case 7:
      cerr << "usrfun seems to be giving incorrect objective derivatives" 
	   << endl;
      break;
    case 8:
      cerr << "usrfun seems to be giving incorrect constraint derivatives" 
	   << endl;
      break;      
    case 9:
      cerr << "the current point cannot be improved" << endl;
      break;
    case 10:
      cerr << "Numerical error in trying to satisfy the linear constraints"
	   << " (or the linearized nonlinear constraints). " 
	   << "The basis is very ill-conditioned" << endl;
      break;
    case 11:
      cerr << "The user signaled undefined functions but no recovery was "
	   << "possible" << endl;
      break;
    case 12:
      cerr << "user signaled to stop using status" << endl;
      break;
    case 20:
      cerr << "Not enough storage for the basis factorization" << endl;
      break;
    case 21:
      cerr << "Error in basis package" << endl;
      break;
    case 22:
      cerr << "The basis is singular after several attempts to factorize it"
	   << "(and add slacks where necessary)" << endl;
      break;
    case 30:
      cerr << "An old basis file had dimensions that did not match the"
	   << " current problem" << endl;
      break;
    case 32:
      cerr << "System error. Wrong number of basic variables" << endl;
      break;
    case 40:
      cerr << "Some input variables have invalid values" << endl;
      break;
    case 41:
      cerr << "The work arrays must have at least 500 elements" << endl;
      break;
    case 42:
      cerr << "Not enough 8-character workspace to solve the problem" << endl;
      break;
    case 43:
      cerr << "Not enough integer workspace to solve the problem" << endl;
      break;
    case 44:
      cerr << "Not enough real workspace to solve the problem" << endl;
      break;
    default:
      cerr << "unrecognized error from inform: " << inform << endl;
    };

    return (inform == 0);
  }

  static void ObjectiveCallback(int *status,
				const int *n,
				const double *x,
				const int *needf,
				const int *nF,
				double *f,
				const int *needG,
				const int *lenG,
				double *G,
				const int *needH,
				const int *lenH,
				double *H,
				double *fMul,
				char *cu,
				const int *lencu,
				int *iu,
				const int *leniu,
				double *ru,
				const int *lenru,
				unsigned int cwLen) {
    assert(1 == *leniu);

    OptProblem *problem = reinterpret_cast<OptProblem*>(iu[0]);

    assert(NULL != problem);
    
    assert(*n == problem->GetNumberOfVariables());
    assert(*nF == problem->GetNumberOfEquations());

    problem->SetState(x);
    
    if (*needf > 0) {
      problem->Evaluate(f);
    }

    assert(*lenG == problem->GetGradientSize());

    if (*needG > 0) {
      problem->Gradient(G);
    }

    *status = 0;
  }

  static void LogCallback(int *status,
			  const int *n,
			  const double *x,
			  const int *needf,
			  const int *nF,
			  double *f,
			  const int *needG,
			  const int *lenG,
			  double *G,
			  const int *needH,
			  const int *lenH,
			  double *H,
			  double *fMul,
			  char *cu,
			  const int *lencu,
			  int *iu,
			  const int *leniu,
			  double *ru,
			  const int *lenru,
			  unsigned int cwLen) {
    *status = 0;
  }
  
  bool Solve(const OptConfig &config,
	     OptProblem *problem,
	     double *x,
	     double *f) {
    
    // clean up workspace from previous calls
    FreeWorkspace();

    // temporarily allocate workspace to init snopt
    AllocateWorkspace(500,500,500);
    
    // init with dummy variables
    Init();

    // now figure out how much space we actually need
    CalculateMemoryRequirements(*problem);
    
    // call init again to make sure that workspace sizes are fixed
    Init();

    // now set configuration options
    SetConfigOptions(config);
    
    // now we're almost ready to do the solve, just need to set up the
    // bajillion or so variables snopt takes as parameters

    // always start "cold". I don't think there's a need to support 
    // warm starts just yet
    int start = 0;

    int neF = problem->GetNumberOfEquations();

    int numberOfVariables = problem->GetNumberOfVariables();
    
    // no support for variable naming
    int nxname = 1;

    // no support for function naming
    int nfname = 1;

    // don't add a constant to the objective function
    double objAdd = 0;

    // the objective function is always in the first row
    int objRow = 1;

    int lenA = problem->GetNumberOfLinearCoefficients();
    // assume all entries of the linear coefficients are specified
    int neA = lenA;
    
    int lenG = problem->GetGradientSize();

    // assume all entries of the gradient are non-empty
    int neG = lenG;

    int lenHu = problem->GetHessianSize();
    
    // assume all entries of the Hessian are non-empty
    int neHu = lenHu;

    // since we're always starting cold initialize the state to be 
    // all zeros
    int *xstate = new int[problem->GetNumberOfVariables()];
    memset(xstate,0,sizeof(int)*problem->GetNumberOfVariables());

    // create the lagrange multiplier array for the variable bound
    // constraints
    double *xmul = new double[problem->GetNumberOfVariables()];
    memset(xmul,0,sizeof(double)*problem->GetNumberOfVariables());
    
    problem->SetState(x);
    problem->Evaluate(f);
        
    // again since we're always starting cold initialize the f state
    int *fstate = new int[problem->GetNumberOfEquations()];
    memset(fstate,0,sizeof(int)*problem->GetNumberOfEquations());
    
    // lagrange multipliers for bounds on the function values
    double *fmul = new double[problem->GetNumberOfEquations()];
    memset(fmul,0,sizeof(double)*problem->GetNumberOfEquations());
    
    // number of superbasic variables... whatever those are
    int ns = 0;

    // number of infeasible variables
    int ninf = 0;

    // sum of infeasible variables
    double sinf = 0;

    int mincw = 0;
    int miniw = 0;
    int minrw = 0;

    // pass a pointer to the problem using the user integer workspace
    int iuser[1];
    iuser[0] = (int)problem;
    int leniuser = 1;

#if 0
    // here's some useful debugging crap
    cerr << "start " << start << endl;
    cerr << "neF " << neF << endl;
    cerr << "numberOfVariables " << numberOfVariables << endl;
    cerr << "nxname " << nxname << endl;
    cerr << "nfname " << nfname << endl;
    cerr << "objAdd " << objAdd << endl;
    cerr << "objRow " << objRow << endl;
    cerr << "problem->GetName().c_str() " << problem->GetName().c_str() << endl;
    for (size_t i = 0; i < neA; ++i) {
      
      cerr << "lin coef f " << i << " " 
	   << problem->GetLinearCoefficientFunctionIndices()[i] << endl;
      cerr << "lin coef v " << i << " " 
	   << problem->GetLinearCoefficientVariableIndices()[i] << endl;
      
    }

    cerr << "lenA " << lenA << endl;
    cerr << "neA " << neA << endl;
    
    for (size_t i = 0; i < neA; ++i) {
      cerr << "lin coef " << i << " " 
	   << problem->GetLinearCoefficients()[i] << endl;
    }
    
    for (size_t i = 0; i < lenG; ++i) {
      cerr << "grad fun " << i << " " 
	   << problem->GetGradientFunctionIndices()[i] << endl;
      cerr << "grad var " << i << " " 
	   << problem->GetGradientVariableIndices()[i] << endl;
    }
    
    cerr << "lenG " << lenG << endl;
    cerr << "neG " << neG << endl;
    
    cerr << "lenHu " << lenHu << endl;
    cerr << "neHu " << neHu << endl;
    
    for (size_t i = 0; i < numberOfVariables; ++i) {
      cerr << "var low bound " << i << " " 
	   << problem->GetVariableLowerBounds()[i] << endl;
      cerr << "var upp bound " << i << " "
	   << problem->GetVariableUpperBounds()[i] << endl;
    }
    
    for (size_t i = 0; i < problem->GetNumberOfEquations(); ++i) {
      cerr << "f low bound " << i << " "
	   << problem->GetEquationLowerBounds()[i] << endl;
      cerr << "f upp bound " << i << " "
	   << problem->GetEquationUpperBounds()[i] << endl;
    }

    cerr << "len cw " << lencw << endl;
    cerr << "len rw " << lenrw << endl;
    cerr << "len iw " << leniw << endl;
#endif
    
    FORTRAN_NAME(snoptf)(&start,
			 &neF,
			 &numberOfVariables,
			 &nxname,
			 &nfname,
			 &objAdd, 
			 &objRow,
			 problem->GetName().c_str(),
			 M_Data::ObjectiveCallback,
			 M_Data::LogCallback,
			 problem->GetLinearCoefficientFunctionIndices(),
			 problem->GetLinearCoefficientVariableIndices(),
			 &lenA, 
			 &neA, 
			 problem->GetLinearCoefficients(),
			 problem->GetGradientFunctionIndices(),
			 problem->GetGradientVariableIndices(),
			 &lenG, 
			 &neG, 
			 problem->GetHessianFunctionIndices(),
			 problem->GetHessianVariableIndices(),
			 &lenHu, 
			 &neHu, 
			 problem->GetVariableLowerBounds(),
			 problem->GetVariableUpperBounds(),
			 "", //empty xnames
			 problem->GetEquationLowerBounds(),
			 problem->GetEquationUpperBounds(),
			 "", //empty fnames
			 x, 
			 xstate, 
			 xmul, 
			 f, 
			 fstate, 
			 fmul, 
			 &inform, 
			 &mincw, 
			 &miniw, 
			 &minrw, 
			 &ns, 
			 &ninf, 
			 &sinf, 
			 cw,
			 &lencw,
			 iuser,
			 &leniuser,
			 rw,
			 &lenrw,
			 cw,
			 &lencw,
			 iw,
			 &leniw,
			 rw,
			 &lenrw,
			 problem->GetName().size(),
			 0, // xname size
			 0, // fname size
			 CWLEN, // cu size
			 CWLEN); // cw size

    bool toReturn = CheckInformAfterSolve();
    
    delete [] xstate;
    delete [] xmul;
    delete [] fstate;
    delete [] fmul;

    return toReturn;
  }

  // file descriptors that snopt uses to dump information
  int iPrint;
  int iSumm;

  // snopt workspace variables
  int lencw;
  int leniw;
  int lenrw;

  char *cw;
  int *iw;
  double *rw;

  // inform keeps track of errors from snopt it should be set to 0 initially
  int inform;
};

OptSnoptSolver::OptSnoptSolver() : m_data(new M_Data())
{

}

OptSnoptSolver::~OptSnoptSolver()
{
  delete m_data;
}

void
OptSnoptSolver::SetPrintFile(int fid)
{
  m_data->iPrint = fid;
}

void
OptSnoptSolver::SetSummaryFile(int fid)
{
  m_data->iSumm = fid;
}

bool
OptSnoptSolver::Solve(const OptConfig &config,
		      OptProblem *problem,
		      double *x,
		      double *f)
{
  return m_data->Solve(config,problem,x,f);
}
